/*-------------------------------------------------------------------
[SCRIPT] Change Tracking Sample   
[DATABASE] 
[DESCRIPTION] 
[MAJ PAR] DATAFLY - Arian Papillon
[DATEMAJ] SQL Saturday - 05/09/2015  
-------------------------------------------------------------------*/
-----------------Prparer l'environnement ------------------------------
-- Base Source
USE master
GO
CREATE DATABASE ChangeTrackingSource
GO
USE ChangeTrackingSource
SELECT  [BusinessEntityID]
      , [PersonType]
      , [FirstName]
      , [LastName]
INTO    person
FROM    AdventureWorks2014.Person.Person 
GO
ALTER TABLE person
ADD CONSTRAINT pk_person PRIMARY KEY CLUSTERED ( BusinessEntityID )
  -- PK ncessaire pour le change tracking
SELECT TOP 1000 * FROM Person
GO
-- Base Destination
USE master
GO
CREATE DATABASE ChangeTrackingTarget
GO
USE ChangeTrackingTarget
GO
CREATE TABLE [SynchroPerson]
    (
      [BusinessEntityID] [INT] NOT NULL
    , [PersonType] [NCHAR](2) NULL
    , [FirstName] NVARCHAR(50) NULL
    , [LastName] NVARCHAR(50) NULL
    , CONSTRAINT [PK_Person_BusinessEntityID] PRIMARY KEY CLUSTERED
        ( [BusinessEntityID] ASC )
    )
GO
INSERT  SynchroPerson
        SELECT  *
        FROM    ChangeTrackingSource..person
 -- mmes donnes
GO


---------------------------------- DEMO  ---------------------------------------------------
-- Activation du change tracking pour la base et pour la table source
USE master
GO
ALTER DATABASE ChangeTrackingSource SET CHANGE_TRACKING= ON
( CHANGE_RETENTION = 2 DAYS, AUTO_CLEANUP = ON
)
GO
USE ChangeTrackingSource
GO
ALTER TABLE person ENABLE CHANGE_TRACKING
WITH (
TRACK_COLUMNS_UPDATED = ON)
GO

-- informations sur le change tracking
SELECT  sc.name AS tracked_schema_name
      , so.name AS tracked_table_name
      , ctt.is_track_columns_updated_on
      , ctt.begin_version /* dbut du CG*/
      , ctt.min_valid_version /*version minimum valide */
      , ctt.cleanup_version /*le cleanup a pu supprimer des donnes jusqu' cette version */
FROM    sys.change_tracking_tables AS ctt
        JOIN sys.objects AS so ON ctt.[object_id] = so.[object_id]
        JOIN sys.schemas AS sc ON so.schema_id = sc.schema_id;
GO

-- tables internes 
SELECT  *
FROM    sys.internal_tables
WHERE   [internal_type_desc] = 'CHANGE_TRACKING'
        OR [internal_type_desc] = 'TRACKED_COMMITTED_TRANSACTIONS'
-- ne peut tre lue qu'avec la DAC
--select * from sys.syscommittab
--select * from sys.change_tracking_245575913

-- Quelle est la version actuelle ?
SELECT   CHANGE_TRACKING_CURRENT_VERSION()
-- Stockons cette valeur
-- Je cre une table vide dans la destination et j'y stocke le n de version de la source
USE ChangeTrackingTarget
CREATE TABLE CT_SourceVersion ( CT_version BIGINT )
INSERT  CT_SourceVersion
        DEFAULT VALUES
USE ChangeTrackingSource
UPDATE  ChangeTrackingTarget..CT_SourceVersion
SET     CT_version = ( SELECT   CHANGE_TRACKING_CURRENT_VERSION())
SELECT  *
FROM    ChangeTrackingTarget..CT_SourceVersion
                     

-- faire quelques mises  jour, ajouts et suppressions successifs dans ma source
UPDATE  person
SET     FirstName = 'Steve'
      , LastName = 'Ballmer'
WHERE   BusinessEntityID = 1

UPDATE  person
SET     FirstName = 'Satya'
      , LastName = 'Nadella'
WHERE   BusinessEntityID = 1

DELETE  person
WHERE   BusinessEntityID = 2

INSERT  person
        ( BusinessEntityID,PersonType,FirstName,LastName )
VALUES  ( 310,'SC', 'Joan','Ballmer')
GO
-- faisons beaucoup de mises  jour
UPDATE  person
SET     FirstName = 'Bill'
      , LastName = 'Gates'
WHERE   BusinessEntityID = 310
GO 5000

-- Voyons ce qui a t track
-- aussi select from change tracking table (DAC)
SELECT  CHANGE_TRACKING_CURRENT_VERSION()
SELECT  CHANGE_TRACKING_MIN_VALID_VERSION(OBJECT_ID('dbo.Person'))
SELECT  *
FROM    CHANGETABLE(CHANGES person, 0) AS CT
-- suivi des colonnes
SELECT  BusinessEntityID
      , SYS_CHANGE_OPERATION
      , CHANGE_TRACKING_IS_COLUMN_IN_MASK(COLUMNPROPERTY(OBJECT_ID('dbo.person'),
                                                         'PersonType',
                                                         'ColumnId'),
                                          CT.SYS_CHANGE_COLUMNS) AS PersonType
      , CHANGE_TRACKING_IS_COLUMN_IN_MASK(COLUMNPROPERTY(OBJECT_ID('dbo.person'),
                                                         'FirstName',
                                                         'ColumnId'),
                                          CT.SYS_CHANGE_COLUMNS) AS FirstName
      , CHANGE_TRACKING_IS_COLUMN_IN_MASK(COLUMNPROPERTY(OBJECT_ID('dbo.person'),
                                                         'LastName',
                                                         'ColumnId'),
                                          CT.SYS_CHANGE_COLUMNS) AS LastName
FROM    CHANGETABLE(CHANGES person, 0) CT
GO

-- informations 
-- combien de transactions ont t faites (toutes tables avec CT) et sont stockes
SELECT  COUNT(*) AS number_commits
      , MIN(commit_time) AS minimum_commit_time
      , MAX(commit_time) AS maximum_commit_time
FROM    sys.dm_tran_commit_table

-- stockage consomm
SELECT  sct1.name AS CT_schema
      , sot1.name AS CT_table
      , ps1.row_count AS CT_rows
      , ps1.reserved_page_count * 8. / 1024. AS CT_reserved_MB
      , sct2.name AS tracked_schema
      , sot2.name AS tracked_name
      , ps2.row_count AS tracked_rows
      , ps2.reserved_page_count * 8. / 1024. AS tracked_base_table_MB
      , CHANGE_TRACKING_MIN_VALID_VERSION(sot2.object_id) AS min_valid_version
FROM    sys.internal_tables it
        JOIN sys.objects sot1 ON it.object_id = sot1.object_id
        JOIN sys.schemas AS sct1 ON sot1.schema_id = sct1.schema_id
        JOIN sys.dm_db_partition_stats ps1 ON it.object_id = ps1.object_id
                                              AND ps1.index_id IN ( 0, 1 )
        LEFT JOIN sys.objects sot2 ON it.parent_object_id = sot2.object_id
        LEFT JOIN sys.schemas AS sct2 ON sot2.schema_id = sct2.schema_id
        LEFT JOIN sys.dm_db_partition_stats ps2 ON sot2.object_id = ps2.object_id
                                                   AND ps2.index_id IN ( 0, 1 )
WHERE   it.internal_type IN ( 209, 210 );


----------- Synchronisons nos tables ----
SELECT  *
FROM    ChangeTrackingSource..Person
WHERE   BusinessEntityID IN ( 1, 2, 310 )
SELECT  *
FROM    ChangeTrackingTarget..SynchroPerson
WHERE   BusinessEntityID IN ( 1, 2, 310 )
   -- les modification ne sont pas dans ma cible

DECLARE @sync_version BIGINT
SELECT  @sync_version = CT_version   -- je rcupre la version stocke
FROM    ChangeTrackingTarget..CT_SourceVersion

IF @sync_version >= CHANGE_TRACKING_MIN_VALID_VERSION(OBJECT_ID('dbo.Person')) -- tester si les donnes sont valides
    BEGIN
-- et j'affiche les modifications de ma source depuis cette version
        SELECT  CT.BusinessEntityID
              , p.PersonType
              , p.FirstName
              , p.LastName
              , CT.SYS_CHANGE_OPERATION
              , CT.SYS_CHANGE_COLUMNS
              , CT.SYS_CHANGE_CONTEXT
        FROM    CHANGETABLE(CHANGES person, @sync_version) CT
                LEFT JOIN person p ON CT.BusinessEntityID = p.BusinessEntityID
    END

-- Je peux donc synchroniser : INSERT, UPDATE, DELETE
DECLARE @sync_version BIGINT
SELECT  @sync_version = CT_version
FROM    ChangeTrackingTarget..CT_SourceVersion

INSERT  ChangeTrackingTarget..SynchroPerson
        ( BusinessEntityID
        , PersonType
        , FirstName
        , LastName
        )
        SELECT  CT.BusinessEntityID
              , p.PersonType
              , p.FirstName
              , p.LastName
        FROM    CHANGETABLE(CHANGES person, @sync_version) CT
                JOIN person p ON CT.BusinessEntityID = p.BusinessEntityID
                                 AND CT.SYS_CHANGE_OPERATION = 'I'

UPDATE  ChangeTrackingTarget..SynchroPerson
SET     FirstName = p.FirstName
      , LastName = p.LastName
      , PersonType = p.PersonType
FROM    ChangeTrackingTarget..SynchroPerson
        JOIN CHANGETABLE(CHANGES person, @sync_version) CT ON ChangeTrackingTarget..SynchroPerson.BusinessEntityID = CT.BusinessEntityID
                                                              AND CT.SYS_CHANGE_OPERATION = 'U'
        JOIN person p ON CT.BusinessEntityID = p.BusinessEntityID

DELETE  ChangeTrackingTarget..SynchroPerson
WHERE   BusinessEntityID IN (
        SELECT  CT.BusinessEntityID
        FROM    CHANGETABLE(CHANGES person, @sync_version) CT
        WHERE   CT.SYS_CHANGE_OPERATION = 'D' )

-- Vrifions
SELECT  *
FROM    ChangeTrackingSource..Person
WHERE   BusinessEntityID IN ( 1, 2, 310 )
SELECT  *
FROM    ChangeTrackingTarget..SynchroPerson
WHERE   BusinessEntityID IN ( 1, 2, 310 )

-- si tout s'est bien pass, je conserve le nouveau numro de version pour la prochaine fois
SELECT   CHANGE_TRACKING_CURRENT_VERSION()
UPDATE  ChangeTrackingTarget..CT_SourceVersion
SET     CT_version = ( SELECT   CHANGE_TRACKING_CURRENT_VERSION()
                     )


